"""Container for the experiment data"""

from collections.abc import Sequence
from typing import Any, Callable, Dict, List, Optional, Union, overload

import pandas as pd

from ml_logger.types import LogType


class Experiment:
    def __init__(
        self,
        config: LogType,
        metrics: Dict[str, pd.DataFrame],
        info: Optional[Dict[Any, Any]] = None,
    ):
        """Class to hold the experiment data

        Args:
            config (LogType): Config used for the experiment
            metrics (Dict[str, pd.DataFrame]): Dictionary mapping strings
                to dataframes. Keys could be "train", "validation", "test"
                and corresponding dataframes would have the data for these
                modes.
            info (Dict[Any, Any]): A dictionary where the user can store
                any information about the experiment (that does not fit
                within config and metrics).
        """
        self.config = config
        self.metrics = metrics
        self.info: Dict[Any, Any] = {}
        if info is not None:
            self.info = info


class ExperimentSequence(Sequence):  # type: ignore
    """Provides a list-like interface to a collection of Experimente
    """

    def __init__(self, experiments: List[Experiment]):
        self.experiments = experiments
        super().__init__()

    def groupby(
        self, group_fn: Callable[[Experiment], str]
    ) -> Dict[str, "ExperimentSequence"]:
        """Group experiments in the sequence

        Args:
            group_fn: Function to assign a string group id to the experiment

        Returns:
            Dict[str, ExperimentSequence]: A dictionary mapping the sring
            group id to a sequence of experiments
        """
        grouped_experiments: Dict[str, List[Experiment]] = {}
        for experiment in self.experiments:
            key = group_fn(experiment)
            if key not in grouped_experiments:
                grouped_experiments[key] = []
            grouped_experiments[key].append(experiment)

        return {
            key: ExperimentSequence(value) for key, value in grouped_experiments.items()
        }

    def filter(self, filter_fn: Callable[[Experiment], bool]) -> "ExperimentSequence":
        """Filter experiments in the sequence

        Args:
            filter_fn: Function to filter an experiment

        Returns:
            ExperimentSequence: A sequence of experiments for which the
            filter condition is true
        """
        return ExperimentSequence(
            [experiment for experiment in self.experiments if filter_fn(experiment)]
        )

    @overload  # noqa:F811
    def __getitem__(self, index: int) -> Experiment:
        pass

    @overload  # noqa:F811
    def __getitem__(self, index: slice) -> "ExperimentSequence":
        pass

    def __getitem__(  # noqa:F811
        self, index: Union[int, slice]
    ) -> Union[Experiment, "ExperimentSequence"]:
        if isinstance(index, slice):
            return ExperimentSequence(self.experiments[index])
        return self.experiments[index]

    def __len__(self) -> int:
        return len(self.experiments)



"""Implementation of Parser to parse experiment from the logs"""

import glob
from typing import Optional

from ml_logger.parser import base as base_parser
from ml_logger.parser.config import (
    parse_json_and_match_key as default_config_line_parser,
)
from ml_logger.parser.experiment.experiment import Experiment
from ml_logger.parser.metric import metrics_to_df
from ml_logger.parser.metric import (
    parse_json_and_match_key as default_metric_line_parser,
)
from ml_logger.types import LogType, ParseLineFunctionType


class Parser(base_parser.Parser):
    """Class to parse an experiment from the log dir"""

    def __init__(
        self,
        parse_config_line: ParseLineFunctionType = default_config_line_parser,
        parse_metric_line: ParseLineFunctionType = default_metric_line_parser,
    ):
        """Class to parse experiment from the logs

        Args:
            parse_config_line (ParseLineFunctionType):
                Function to parse a config line in the log file. The function
                should return None if the line is not a valid config log
                (eg error messages)
            parse_metric_line (ParseLineFunctionType):
                Function to parse a metric line in the log file. The function
                should return None if the line is not a valid metric log
                (eg error messages)
        """
        self.log_key = "logbook_type"
        self.log_type = "experiment"
        self.parse_line = self._wrap_parse_line(parse_config_line, parse_metric_line)

    def _wrap_parse_line(
        self,
        parse_config_line: ParseLineFunctionType,
        parse_metric_line: ParseLineFunctionType,
    ) -> ParseLineFunctionType:
        def fn(line: str) -> Optional[LogType]:
            log = parse_config_line(line)
            if log is not None:
                if self.log_key not in log:
                    log[self.log_key] = "config"
            else:
                log = parse_metric_line(line)
                if log is not None and self.log_key not in log:
                    log[self.log_key] = "metric"
            return log

        return fn

    def parse(self, filepath_pattern: str) -> Experiment:
        """Load one experiment from the log dir

        Args:
            filepath_pattern (str): filepath pattern to glob
        Returns:
            Experiment
        """
        configs = []
        metric_logs = []
        paths = glob.glob(filepath_pattern)
        for file_path in paths:
            for log in self._parse_file(file_path=file_path):
                # At this point, if log is not None, it will have a key self.log_key
                if log is not None:
                    if log[self.log_key] == "config":
                        configs.append(log)
                    elif log[self.log_key] == "metric":
                        metric_logs.append(log)
        return Experiment(
            config=configs[-1], metrics=metrics_to_df(metric_logs=metric_logs), info={},
        )



"""Logger class that writes to wandb"""

import wandb

from ml_logger.logger.base import Logger as BaseLogger
from ml_logger.types import ConfigType, LogType, MetricType


class Logger(BaseLogger):
    """Logger class that writes to wandb
    """

    def __init__(self, config: ConfigType):
        """Initialise the Wandb Logger

        Args:
            config (ConfigType): config to initialise the wandb logger.
                The config can have any parameters that wandb.init() method
                accepts (https://docs.wandb.com/library/init). Note that
                the config is passed as keyword arguments to the wandb.init()
                method. This provides a lot of flexibility to the users to
                configure wandb. This also means that config should not
                have any parameters that wandb.init() would not accept.
        """
        super().__init__(config=config)
        self.keys_to_skip = ["logbook_id", "logbook_type", "logbook_timestamp"]
        self.keys_to_check = ["step"]
        self.run = wandb.init(**config)

    def write_log(self, log: LogType) -> None:
        """Write the log to wandb

        Args:
            log (LogType): Log to write
        """
        logbook_type = log["logbook_type"]
        if logbook_type == "metric":
            log = self._prepare_metric_log_to_write(log=log)
            self.write_metric_log(metric=log)
        else:
            log = self._prepare_log_to_write(log=log)
            if logbook_type == "config":
                self.write_config(config=log)
            # Only metric logs and message logs are supported right now

    def write_metric_log(self, metric: MetricType) -> None:
        """Write metric to wandb

        Args:
            metric (MetricType): Metric to write
        """
        for key in self.keys_to_check:  # type: ignore
            assert key in metric
        step = metric.pop("step")
        if self.key_prefix:
            prefix = {metric.pop(self.key_prefix)}
            metric = {f"{prefix}_{key}": value for key, value in metric.items()}
        wandb.log(metric, step)

    def write_config(self, config: ConfigType) -> None:
        """Write the config to wandb

        Args:
            config (ConfigType): Config to write
        """
        wandb.config.update(config)



"""Functions to interface with the filesystem"""

import json
import logging
import os
from functools import partial
from typing import Optional

from ml_logger.logger.base import Logger as BaseLogger
from ml_logger.types import ConfigType, LogType
from ml_logger.utils import make_dir


def _serialize_log_to_json(log: LogType) -> str:
    """Serialize the log into a JSON string

    Args:
        log (LogType): Log to be serialized

    Returns:
        str: JSON serialized string
    """
    return json.dumps(log)


def _get_logger(logger_name: str = "default_logger") -> logging.Logger:
    """Get logger for a given name

    Args:
        logger_name (str, optional): Name of the logger (to retrieve).
            Defaults to "default_logger"

    Returns:
        logging.Logger: Logger object
    """
    return logging.getLogger(name=logger_name)


def _set_logger(
    logger_file_path: str,
    logger_name: str = "default_logger",
    write_to_console: bool = True,
) -> logging.Logger:
    """Set logger to log to the given path

    Modified from https://docs.python.org/3/howto/logging-cookbook.html

    Args:
        logger_file_path (str): Filepath to write to
        logger_name (str, optional): Name of the logger to use. Defaults
            to "default_logger"
        write_to_console (bool, optional): Should write the logs to console.
            Defaults to True

    Returns:
        logging.Logger: Logger object
    """
    logger = logging.getLogger(name=logger_name)
    logger.setLevel(level=logging.INFO)
    # create file handler which logs all the messages
    file_handler = logging.FileHandler(filename=logger_file_path)
    file_handler.setLevel(level=logging.INFO)
    # create formatter and add it to the handlers
    formatter = logging.Formatter(fmt="%(message)s")
    file_handler.setFormatter(fmt=formatter)
    # add the handlers to the logger
    logger.addHandler(hdlr=file_handler)

    if write_to_console:
        # create console handler with a higher log level
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(level=logging.INFO)
        # add formatter to the handlers
        stream_handler.setFormatter(fmt=formatter)
        # add the handlers to the logger
        logger.addHandler(hdlr=stream_handler)
    return logger


class Logger(BaseLogger):
    """Logger class that writes to the filesystem
    """

    def __init__(self, config: ConfigType):
        """Initialise the Filesystem Logger

        Args:
            config (ConfigType): config to initialise the filesystem logger.
                It must have two keys: logger_file_path and logger_name.
                "logger_file_path" is the path to the file where the logs
                will be written. "logger_name" is the name of the logger instance
        """
        super().__init__(config=config)
        assert "logger_dir" in config
        assert "logger_name" in config
        assert "write_to_console" in config
        assert "create_multiple_log_files" in config
        assert "filename_prefix" in config
        assert "filename" in config

        logger_types = ["config", "message", "metadata", "metric"]

        make_dir(config["logger_dir"])

        _get_logger_file_path = partial(
            get_logger_file_path,
            logger_dir=config["logger_dir"],
            filename=config["filename"],
            filename_prefix=config["filename_prefix"],
        )

        if config["create_multiple_log_files"]:

            self.loggers = {
                _type: _set_logger(
                    logger_file_path=_get_logger_file_path(
                        filename_suffix=f"{_type}_log"
                    ),
                    logger_name=config["logger_name"] + "_" + _type,
                    write_to_console=config["write_to_console"],
                )
                for _type in logger_types
            }

        else:
            logger = _set_logger(
                logger_file_path=_get_logger_file_path(filename_suffix="log"),
                logger_name=config["logger_name"],
                write_to_console=config["write_to_console"],
            )
            self.loggers = {_type: logger for _type in logger_types}

    def write_log(self, log: LogType) -> None:
        """Write the log to the filesystem

        Args:
            log (LogType): Log to write
        """

        log_str = _serialize_log_to_json(log=self._prepare_log_to_write(log))
        return self._write_log_to_fs(log_str=log_str, log_type=log["logbook_type"])

    def _write_log_to_fs(self, log_str: str, log_type: str) -> None:
        """Write log string to filesystem

        Args:
            log_str (str): Log string to write
            log_type (str): Type of log to write
        """
        if log_type not in self.loggers:
            log_type = "message"
        self.loggers[log_type].info(msg=log_str)


def get_logger_file_path(
    logger_dir: str, filename: Optional[str], filename_prefix: str, filename_suffix: str
) -> str:
    if filename is None:
        filename = f"{filename_prefix}{filename_suffix}.jsonl"
    logger_file_path = os.path.join(logger_dir, filename)
    return logger_file_path



"""Logger class that writes to tensorboard"""

from typing import Dict

from tensorboardX import SummaryWriter

from ml_logger.logger.base import Logger as BaseLogger
from ml_logger.types import ConfigType, LogType, MetricType, NumType
from ml_logger.utils import make_dir


class Logger(BaseLogger):
    """Logger class that writes to tensorboardX
    """

    def __init__(self, config: ConfigType):
        """Initialise the tensorboardX Logger

        Args:
            config (ConfigType): config to initialise the tensorboardX
                logger. The config can have any parameters that
                tensorboardX.SummaryWriter() method accepts
                (https://tensorboardx.readthedocs.io/en/latest/tensorboard.html#tensorboardX.SummaryWriter).
                Note that the config is passed as keyword arguments to the
                tensorboardX.SummaryWriter() method. This provides a lot
                of flexibility to the users to configure tensorboard. This also
                means that config should not have any parameters that
                tensorboardX.SummaryWriter() would not accept.
        """
        super().__init__(config=config)
        make_dir(config["logdir"])
        self.summary_writer = SummaryWriter(**config)
        self.keys_to_skip = ["logbook_id", "logbook_type", "logbook_timestamp"]

    def write_log(self, log: LogType) -> None:
        """Write the log to tensorboard

        Args:
            log (LogType): Log to write
        """
        logbook_type = log["logbook_type"]
        if logbook_type == "metric":
            log = self._prepare_metric_log_to_write(log=log)
            self.write_metric_log(metric=log)
        else:
            if logbook_type == "config":
                self.write_config(config=log)
            # Only metric logs and message logs are supported right now

    def write_metric_log(self, metric: MetricType) -> None:
        """Write metric to tensorboard

        Args:
            metric (MetricType): Metric to write
        """
        global_step = None
        if "global_step" in metric:
            global_step = metric.pop("global_step")
        walltime = None
        if "walltime" in metric:
            walltime = metric.pop("walltime")

        main_tag = ""
        if "tag" in metric:
            main_tag = str(metric.pop("tag")) + "/"
        elif "main_tag" in metric:
            main_tag = str(metric.pop("main_tag")) + "/"

        if self.key_prefix:
            prefix = {metric.pop(self.key_prefix)}
            metric = {f"{prefix}_{key}": value for key, value in metric.items()}

        for key, value in metric.items():
            self.summary_writer.add_scalar(
                tag=f"{main_tag}{key}",
                scalar_value=value,
                global_step=global_step,
                walltime=walltime,
            )

    def write_config(self, config: ConfigType) -> None:
        """Write the config to tensorboard

        Args:
            config (ConfigType): Config to write
        """
        name = None
        if "name" in config:
            name = config.pop("name")

        metric_dict: Dict[str, NumType] = {}
        if "metric_dict" in config:
            metric_dict = config.pop("metric_dict")
            metric_dict = self._prepare_metric_log_to_write(log=metric_dict)

        global_step = None
        if "global_step" in config:
            global_step = config.pop("global_step")

        config = self._prepare_log_to_write(log=config)

        for key in config:
            if config[key] is None:
                config[key] = "None"

        self.summary_writer.add_hparams(
            hparam_dict=config,
            metric_dict=metric_dict,
            name=name,
            global_step=global_step,
        )



from ml_logger.parser.experiment.experiment import Experiment  # noqa:F401
from ml_logger.parser.experiment.experiment import ExperimentSequence  # noqa:F401
from ml_logger.parser.experiment.parser import Parser  # noqa:F401



"""Implementation of Parser to parse metrics from logs"""

from typing import Callable, Dict, List, Optional

import pandas as pd

from ml_logger.parser import log as log_parser
from ml_logger.parser import utils as parser_utils
from ml_logger.parser.utils import parse_json
from ml_logger.types import LogType, MetricType, ParseLineFunctionType


def parse_json_and_match_key(line: str) -> Optional[LogType]:
    log = parse_json(line)
    if log:
        key = "logbook_type"
        if key not in log or log[key] != "metric":
            log = None
    return log


def group_metrics(metrics: List[MetricType]) -> Dict[str, List[MetricType]]:
    """Group a list of metrics into a dictionary of (key,
    list of grouped metrics)

    Args:
        metrics (List[MetricType]): List of metrics to group

    Returns:
        Dict[str, List[MetricType]]: Dictionary of (key,
            list of grouped metrics)
    """
    return {"all": metrics}


def aggregate_metrics(metrics: List[MetricType]) -> List[MetricType]:
    """Aggregate a list of metrics

    Args:
        metrics (List[MetricType]): List of metrics to aggregate

    Returns:
        List[MetricType]: List of aggregated metrics
    """
    return metrics


class Parser(log_parser.Parser):
    """Class to parse the metrics from the logs"""

    def __init__(self, parse_line: ParseLineFunctionType = parse_json_and_match_key):
        """Class to parse the metrics from the logs

        Args:
            parse_line (ParseLineFunctionType):
                Function to parse a line in the log file. The function
                should return None if the line is not a valid log statement
                (eg error messages). Defaults to parse_json_and_match_key.
        """
        super().__init__(parse_line)
        self.log_type = "metric"

    def parse_as_df(
        self,
        filepath_pattern: str,
        group_metrics: Callable[
            [List[LogType]], Dict[str, List[LogType]]
        ] = group_metrics,
        aggregate_metrics: Callable[[List[LogType]], List[LogType]] = aggregate_metrics,
    ) -> Dict[str, pd.DataFrame]:
        """Create a dict of (metric_name, dataframe)

        Method that:
        (i) reads metrics from the filesystem
        (ii) groups metrics
        (iii) aggregates all the metrics within a group,
        (iv) converts the aggregate metrics into dataframes and returns a \
            dictionary of dataframes

        Args:
            filepath_pattern (str): filepath pattern to glob
            group_metrics (Callable[[List[LogType]], Dict[str, List[LogType]]], optional):
                Function to group a list of metrics into a dictionary of
                (key, list of grouped metrics). Defaults to group_metrics.
            aggregate_metrics (Callable[[List[LogType]], List[LogType]], optional):
                Function to aggregate a list of metrics. Defaults to aggregate_metrics.

        """
        metric_logs = list(self.parse(filepath_pattern))
        return metrics_to_df(
            metric_logs=metric_logs,
            group_metrics=group_metrics,
            aggregate_metrics=aggregate_metrics,
        )


def metrics_to_df(
    metric_logs: List[LogType],
    group_metrics: Callable[[List[LogType]], Dict[str, List[LogType]]] = group_metrics,
    aggregate_metrics: Callable[[List[LogType]], List[LogType]] = aggregate_metrics,
) -> Dict[str, pd.DataFrame]:
    """Create a dict of (metric_name, dataframe)

    Method that:
    (i) groups metrics
    (ii) aggregates all the metrics within a group,
    (iii) converts the aggregate metrics into dataframes and returns a \
        dictionary of dataframes

    Args:
        metric_logs (List[LogType]): List of metrics
        group_metrics (Callable[[List[LogType]], Dict[str, List[LogType]]], optional):
            Function to group a list of metrics into a dictionary of
            (key, list of grouped metrics). Defaults to group_metrics.
        aggregate_metrics (Callable[[List[LogType]], List[LogType]], optional):
            Function to aggregate a list of metrics. Defaults to aggregate_metrics.

    Returns:
        Dict[str, pd.DataFrame]: [description]

    """
    grouped_metrics: Dict[str, List[LogType]] = group_metrics(metric_logs)
    aggregated_metrics = {
        key: aggregate_metrics(metrics) for key, metrics in grouped_metrics.items()
    }

    metric_dfs = {
        key: pd.DataFrame.from_dict(
            data=parser_utils.map_list_of_dicts_to_dict_of_lists(metrics)
        )
        for key, metrics in aggregated_metrics.items()
    }
    return metric_dfs



"""Implementation of Parser to parse config from logs"""

from typing import Optional

from ml_logger.parser import log as log_parser
from ml_logger.parser.utils import parse_json
from ml_logger.types import LogType, ParseLineFunctionType


def parse_json_and_match_key(line: str) -> Optional[LogType]:
    log = parse_json(line)
    if log:
        key = "logbook_type"
        if key not in log or log[key] != "config":
            log = None
    return log


class Parser(log_parser.Parser):
    """Class to parse config from the logs"""

    def __init__(self, parse_line: ParseLineFunctionType = parse_json_and_match_key):
        """Class to parse config from the log

        Args:
            parse_line (ParseLineFunctionType):
                Function to parse a line in the log file. The function
                 should return None if the line is not a valid config (eg
                 error messages). Defaults to parse_json_and_match_key.
        """
        super().__init__(parse_line=parse_line)
        self.log_type = "config"



from typing import Any, Callable, Dict, Optional, Union

NumType = Union[int, float]
ValueType = Union[str, int, float]
LogType = Dict[str, Any]
ParseLineFunctionType = Callable[[str], Optional[LogType]]
ConfigType = LogType
MetricType = Dict[str, ValueType]
ModelType = Any
ComparisonOpType = Callable[[ValueType, ValueType], bool]
KeyMapType = Dict[str, str]



"""Implementation of Parser to parse the logs"""

import glob
from typing import Iterator, Optional

from ml_logger.parser.base import Parser as BaseParser
from ml_logger.parser.utils import parse_json
from ml_logger.types import LogType, ParseLineFunctionType


class Parser(BaseParser):
    """Class to parse the log files"""

    def __init__(self, parse_line: ParseLineFunctionType = parse_json):
        """Class to parse the log files

        Args:
            parse_line (ParseLineFunctionType):
                Function to parse a line in the log file. The function
                should return None if the line is not a valid log statement
                (eg error messages). Defaults to parse_json.
        """
        super().__init__(parse_line)
        self.log_type = "log"
        self.parse_line = self._wrap_parse_line(parse_line)

    def _wrap_parse_line(
        self, parse_line: ParseLineFunctionType
    ) -> ParseLineFunctionType:
        def fn(line: str) -> Optional[LogType]:
            log = parse_line(line)
            if log is not None and self.log_key not in log:
                log[self.log_key] = self.log_type
            return log

        return fn

    def _parse_file(self, file_path: str) -> Iterator[Optional[LogType]]:
        """Open a log file and parse its content

        Args:
            file_path (str): Log file to read from

        Returns:
            Iterator[Optional[LogType]]: Iterator over the logs

        Yields:
            Iterator[Optional[LogType]]: Iterator over the logs
        """

        with open(file_path) as f:
            for line in f:
                log = self.parse_line(line)
                yield log

    def parse(self, filepath_pattern: str) -> Iterator[LogType]:
        """Open a log file, parse its contents and return `logs`

        Args:
            filepath_pattern (str): filepath pattern to glob

        Returns:
            Iterator[LogType]: Iterator over the logs

        Yields:
            Iterator[LogType]: Iterator over the logs
        """
        paths = glob.iglob(filepath_pattern)
        for file_path in paths:
            for log in self._parse_file(file_path):
                if log is not None:
                    yield log

    def parse_first_log(self, filepath_pattern: str) -> Optional[LogType]:
        """Return the first log from a file. The method will return after
        finding the first log. Unlike `get_logs()` method, it will not
        iterate over the entire log file (thus saving memory and time).

        Args:
            filepath_pattern (str): filepath pattern to glob

        Returns:
            LogType: First instance of a log

        """
        paths = glob.iglob(filepath_pattern)
        for file_path in paths:
            for log in self._parse_file(file_path):
                if log is not None:
                    return log
        return None

    def parse_last_log(self, filepath_pattern: str) -> Optional[LogType]:
        """Return the last log from a file. Like `get_logs()` method,
        it will iterate over the entire log file but will not keep all
        the logs in memory (thus saving memory).

        Args:
            filepath_pattern (str): filepath pattern to glob

        Returns:
            LogType: Last instance of a log

        """
        last_log: Optional[LogType] = None
        paths = glob.iglob(filepath_pattern)
        for file_path in paths:
            for log in self._parse_file(file_path=file_path):
                if log is not None:
                    last_log = log
        return last_log



"""Base class that all parsers extend"""

from abc import ABC
from typing import Iterator, Optional

from ml_logger.parser.utils import parse_json
from ml_logger.types import LogType, ParseLineFunctionType


class Parser(ABC):
    """Base class that all parsers extend"""

    def __init__(self, parse_line: ParseLineFunctionType = parse_json):
        """Class to parse the log files

        Args:
            parse_line (ParseLineFunctionType):
                Function to parse a line in the log file. The function
                should return None if the line is not a valid log statement
                (eg error messages). Defaults to parse_json.
        """
        self.log_key = "logbook_type"
        self.log_type = "base_parser"
        self.parse_line = parse_line

    def _parse_file(self, file_path: str) -> Iterator[Optional[LogType]]:
        """Open a log file and parse its content

        Args:
            file_path (str): Log file to read from

        Returns:
            Iterator[Optional[LogType]]: Iterator over the logs

        Yields:
            Iterator[Optional[LogType]]: Iterator over the logs
        """

        with open(file_path) as f:
            for line in f:
                log = self.parse_line(line)
                yield log
